use std::{fmt::Display, time::Duration};

use ev3dev_lang_rust::Screen;
use image::Rgb;
use imageproc::{
    drawing::{draw_line_segment_mut, draw_polygon_mut, draw_text_mut, text_size},
    point::Point,
};
use itertools::Itertools;
use rusttype::{Font, Scale};

const BACKGROUND: Rgb<u8> = Rgb([255; 3]);
const FOREGROUND: Rgb<u8> = Rgb([0; 3]);
const FONT_DATA: &[u8] = include_bytes!("../fonts/RobotoMono-Regular.ttf");

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
pub enum Parameter {
    Kp,
    Ki,
    Kd,
    Speed,
}

#[allow(clippy::similar_names)]
pub fn draw_debug<'a>(screen: &mut Screen, text: impl IntoIterator<Item = &'a str>, font: &Font) {
    #[allow(clippy::cast_precision_loss)]
    let scale = Scale {
        x: screen.xres() as f32 / 10.0,
        y: screen.yres() as f32 / 4.0 - 8.0,
    };
    let xres = i32::try_from(screen.xres()).expect("range checked");
    let yres = i32::try_from(screen.yres()).expect("range checked");
    let center = xres / 2;
    clear(screen);
    for (i, line) in text.into_iter().enumerate().take(4) {
        draw_centered_text(
            screen,
            center,
            yres * i32::try_from(i).expect("there is only place for 4 lines") / 4 + 4,
            scale,
            font,
            line,
        );
    }
    screen.update();
}

#[allow(clippy::similar_names)]
pub fn draw_calibration(screen: &mut Screen, parameter: &str, value: impl Display, font: &Font) {
    #[allow(clippy::cast_precision_loss)]
    let smal_scale = Scale::uniform(screen.yres() as f32 / 4.0 - 8.0);
    #[allow(clippy::cast_precision_loss)]
    let big_scale = Scale::uniform(screen.yres() as f32 / 2.0 - 8.0);
    let xres = i32::try_from(screen.xres()).expect("range checked");
    let yres = i32::try_from(screen.yres()).expect("range checked");
    let center = xres / 2;
    clear(screen);
    draw_centered_text(screen, center, 4, smal_scale, font, "kalibrieren:");
    draw_centered_text(screen, center, yres / 4 + 4, smal_scale, font, parameter);
    draw_centered_text(
        screen,
        center,
        yres / 2 + 4,
        big_scale,
        font,
        format!("{value:.2}").as_str(),
    );
    screen.update();
}

pub fn draw_settings(
    screen: &mut Screen,
    kp: impl Display,
    ki: impl Display,
    kd: impl Display,
    v: impl Display,
    selected: Parameter,
    font: &Font,
) {
    #[allow(clippy::cast_precision_loss)]
    let text_scale = Scale::uniform(screen.yres() as f32 / 4.0 - 8.0);
    clear(screen);
    draw_centered_text(
        screen,
        i32::try_from(screen.xres()).expect("range checked") / 2,
        4,
        text_scale,
        font,
        "Settings",
    );

    draw_setting(screen, kp, selected == Parameter::Kp, 0, "k_p", font);
    draw_setting(screen, ki, selected == Parameter::Ki, 1, "k_i", font);
    draw_setting(screen, kd, selected == Parameter::Kd, 2, "k_d", font);
    draw_setting(screen, v, selected == Parameter::Speed, 3, "v", font);

    screen.update();
}

pub fn draw_driving(screen: &mut Screen, font: &Font) {
    clear(screen);
    draw_centered_text(
        screen,
        i32::try_from(screen.xres()).expect("range checked") / 2,
        4,
        #[allow(clippy::cast_precision_loss)]
        Scale::uniform(screen.yres() as f32 / 2.0 - 8.0),
        font,
        "Fahrt",
    );
    screen.update();
}

#[allow(clippy::similar_names)]
pub fn draw_finished(
    screen: &mut Screen,
    time: Duration,
    (k_p, k_i, k_d): (f32, f32, f32),
    v: i32,
    font: &Font,
) {
    clear(screen);
    #[allow(clippy::cast_precision_loss)]
    let text_scale = Scale::uniform(screen.yres() as f32 / 4.0 - 8.0);
    #[allow(clippy::cast_precision_loss)]
    let time_scale = Scale::uniform(screen.yres() as f32 / 2.0 - 8.0);
    #[allow(clippy::cast_precision_loss)]
    let value_scale = Scale {
        x: screen.xres() as f32 / 10.0,
        y: screen.yres() as f32 / 4.0 - 8.0,
    };
    let xres = i32::try_from(screen.xres()).expect("range checked");
    let yres = i32::try_from(screen.yres()).expect("range checked");
    draw_centered_text(screen, xres / 2, 4, text_scale, font, "Zeit:");
    draw_centered_text(
        screen,
        xres / 2,
        yres / 4 + 4,
        time_scale,
        font,
        format!("{:.2}s", time.as_secs_f32()).as_str(),
    );
    macro_rules! draw_param {
        ($param: ident, $index: literal) => {
            draw_centered_text(
                screen,
                xres * (1 + 2 * $index) / 8,
                yres * 3 / 4 + 4,
                value_scale,
                font,
                format!("{:.2}", $param).as_str(),
            );
        };
    }
    draw_param!(k_p, 0);
    draw_param!(k_i, 1);
    draw_param!(k_d, 2);
    draw_param!(v, 3);
    screen.update();
}

pub fn font() -> Font<'static> {
    Font::try_from_bytes(FONT_DATA).unwrap()
}

#[allow(clippy::similar_names)]
fn draw_setting(
    screen: &mut Screen,
    value: impl Display,
    selected: bool,
    index: i32,
    name: &str,
    font: &Font,
) {
    #[allow(clippy::cast_precision_loss)]
    let text_scale = Scale::uniform(screen.yres() as f32 / 4.0 - 8.0);
    #[allow(clippy::cast_precision_loss)]
    let value_scale = Scale {
        x: screen.xres() as f32 / 10.0,
        y: screen.yres() as f32 / 4.0 - 8.0,
    };
    let xres = i32::try_from(screen.xres()).expect("range checked");
    let yres = i32::try_from(screen.yres()).expect("range checked");
    let x = xres * (1 + 2 * index) / 8;
    draw_centered_text(screen, x, yres / 4 + 4, text_scale, font, name);
    draw_centered_text(
        screen,
        x,
        yres * 5 / 8 + 4,
        value_scale,
        font,
        format!("{value:.2}").as_str(),
    );
    let top_triangle = [
        Point::new(x, yres / 2),
        Point::new(x - 10, yres / 2 + 10),
        Point::new(x + 10, yres / 2 + 10),
    ];
    let bottom_triangle = [
        Point::new(x, yres),
        Point::new(x - 10, yres - 10),
        Point::new(x + 10, yres - 10),
    ];
    if selected {
        draw_polygon_mut(&mut screen.image, &top_triangle, FOREGROUND);
        draw_polygon_mut(&mut screen.image, &bottom_triangle, FOREGROUND);
    } else {
        for (start, end) in top_triangle
            .into_iter()
            .circular_tuple_windows()
            .chain(bottom_triangle.into_iter().circular_tuple_windows())
        {
            draw_line_segment_mut(
                &mut screen.image,
                point2tuple(start),
                point2tuple(end),
                FOREGROUND,
            );
        }
    }
}

fn draw_centered_text(
    screen: &mut ev3dev_lang_rust::Screen,
    x: i32,
    y: i32,
    scale: Scale,
    font: &Font,
    text: &str,
) {
    let (width, _) = text_size(scale, font, text);
    draw_text_mut(
        &mut screen.image,
        FOREGROUND,
        x - width / 2,
        y,
        scale,
        font,
        text,
    );
}

fn clear(screen: &mut Screen) {
    screen.image.fill(BACKGROUND.0[0]);
}

const fn point2tuple(point: Point<i32>) -> (f32, f32) {
    #[allow(clippy::cast_precision_loss)]
    (point.x as f32, point.y as f32)
}
