mod buttons;
mod display;

use buttons::SmartEv3Buttons;
use display::{
    draw_calibration, draw_debug, draw_driving, draw_finished, draw_settings, font, Parameter,
};
use ev3dev_lang_rust::{
    motors::LargeMotor, sensors::ColorSensor, Device, Ev3Error, Ev3Result, Screen,
};
use pid::Pid;
use rand::{thread_rng, Rng};
use rusttype::Font;
use std::{
    error::Error,
    iter,
    thread::sleep,
    time::{Duration, Instant},
};
use thiserror::Error;

fn main() {
    while try_main().is_err() {
        sleep(Duration::from_millis(10));
    }
}

fn try_main() -> Result<(), Box<dyn Error>> {
    let font = font();
    let mut screen = Screen::new()?;
    let motors = LargeMotor::list()?;
    let [right_motor, left_motor] = &motors[..] else {
        let motor_texts: Vec<_> = motors.iter().map(|motor| format!("{:?}", motor.get_driver_name())).collect();
        draw_debug(&mut screen, iter::once("motors:").chain(motor_texts.iter().map(String::as_str)), &font);
        Err("need two motors")?
    };
    let sensor = ColorSensor::find().map_err(|e| {
        draw_debug(&mut screen, ["no color sensor"], &font);
        e
    })?;
    let mut buttons = SmartEv3Buttons::new()?;

    left_motor.set_polarity(LargeMotor::POLARITY_INVERSED)?;
    right_motor.set_polarity(LargeMotor::POLARITY_INVERSED)?;
    left_motor.set_stop_action(LargeMotor::STOP_ACTION_BRAKE)?;
    right_motor.set_stop_action(LargeMotor::STOP_ACTION_BRAKE)?;

    sensor.set_mode_rgb_raw()?;

    let black = calibrate_gray(&mut screen, &sensor, &mut buttons, "Schwarz", &font)?;
    let white = calibrate_gray(&mut screen, &sensor, &mut buttons, "Weiß", &font)?;
    let setpoint = (black + white) / 2.0;

    loop {
        let (k_p, k_i, k_d, v) = select_values(&mut screen, &mut buttons, &font);

        draw_driving(&mut screen, &font);
        let result = follow_line(
            (left_motor, right_motor),
            &sensor,
            &mut buttons,
            setpoint,
            (k_p / 10.0, k_i / 10000.0, k_d / 10000.0),
            v,
        );
        left_motor.stop().and(right_motor.stop())?;
        let time = match result {
            Ok(time) => time,
            Err(ProgramError::UserAbort) => continue,
            e => e?,
        };

        draw_finished(&mut screen, time, (k_p, k_i, k_d), v, &font);
        buttons.process();
        while !buttons.is_enter_pressed() {
            buttons.process();
        }
    }
}

fn select_values(
    screen: &mut Screen,
    buttons: &mut SmartEv3Buttons,
    font: &Font,
) -> (f32, f32, f32, i32) {
    use Parameter::{Kd, Ki, Kp, Speed};
    let mut k_p = thread_rng().gen_range(0.01..=10.0);
    let mut k_i = thread_rng().gen_range(0.01..=10.0);
    let mut k_d = thread_rng().gen_range(0.01..=10.0);
    let mut v = thread_rng().gen_range(1..=100);
    let mut selected = Kp;
    draw_settings(screen, k_p, k_i, k_d, v, selected, font);

    loop {
        buttons.process();
        if buttons.is_enter_pressed() {
            break;
        }
        if buttons.is_right_pressed() {
            selected = match selected {
                Kp => Ki,
                Ki => Kd,
                Kd => Speed,
                Speed => Kp,
            };
        }
        if buttons.is_left_pressed() {
            selected = match selected {
                Kp => Speed,
                Ki => Kp,
                Kd => Ki,
                Speed => Kd,
            };
        }
        match selected {
            Speed => {
                if buttons.is_up() {
                    v = (v + 1).min(100);
                }
                if buttons.is_down() {
                    v = (v - 1).max(0);
                }
            }
            p => {
                let param = match p {
                    Kp => &mut k_p,
                    Ki => &mut k_i,
                    Kd => &mut k_d,
                    Speed => unreachable!(),
                };
                if buttons.is_up() {
                    *param = (*param
                        + if buttons.is_up_pressed() {
                            0.01f32
                        } else {
                            0.1f32
                        })
                    .min(99.99);
                }
                if buttons.is_down() {
                    *param = (*param
                        - if buttons.is_down_pressed() {
                            0.01f32
                        } else {
                            0.1f32
                        })
                    .max(0.0);
                }
            }
        }
        if buttons.is_left_pressed()
            || buttons.is_right_pressed()
            || buttons.is_up()
            || buttons.is_down()
        {
            draw_settings(screen, k_p, k_i, k_d, v, selected, font);
        }
        if buttons.is_up_pressed() || buttons.is_down_pressed() {
            sleep(Duration::from_millis(200));
        } else {
            sleep(Duration::from_millis(10));
        }
    }

    (k_p, k_i, k_d, v)
}

fn calibrate_gray(
    screen: &mut Screen,
    sensor: &ColorSensor,
    buttons: &mut SmartEv3Buttons,
    name: &str,
    font: &Font,
) -> Ev3Result<f32> {
    loop {
        let measurement = gray(sensor.get_rgb()?);
        draw_calibration(screen, name, measurement, font);
        buttons.process();
        if buttons.is_enter_pressed() {
            break Ok(measurement);
        }
    }
}

#[derive(Error, Debug)]
enum ProgramError {
    #[error("working with EV3 peripheral")]
    Ev3(#[from] Ev3Error),
    #[error("the user abortet the run")]
    UserAbort,
}

enum LineFollowState {
    PreStart,
    Start(Instant),
    Run(Instant),
    Finish(Duration),
}

fn follow_line(
    (left, right): (&LargeMotor, &LargeMotor),
    sensor: &ColorSensor,
    buttons: &mut SmartEv3Buttons,
    setpoint: f32,
    (k_p, k_i, k_d): (f32, f32, f32),
    v: i32,
) -> Result<Duration, ProgramError> {
    let mut controller = Pid::new(setpoint, 200.0);
    controller
        .p(k_p, f32::INFINITY)
        .i(k_i, f32::INFINITY)
        .d(k_d, f32::INFINITY);
    left.set_duty_cycle_sp(0)?;
    right.set_duty_cycle_sp(0)?;
    buttons.process();
    let mut state = LineFollowState::PreStart;

    left.run_direct()?;
    right.run_direct()?;
    let mut next = Instant::now();
    while !(buttons.is_enter_pressed() || matches!(state, LineFollowState::Finish(_))) {
        let color = sensor.get_rgb()?;
        let measurement = gray(color);
        #[allow(clippy::cast_possible_truncation)]
        let controll = controller.next_control_output(measurement).output as i32;
        left.set_duty_cycle_sp((v + controll).clamp(-100, 100))?;
        right.set_duty_cycle_sp((v - controll).clamp(-100, 100))?;
        let red = is_red(color);
        match state {
            LineFollowState::PreStart if red => state = LineFollowState::Start(Instant::now()),
            LineFollowState::Start(start) if !red => state = LineFollowState::Run(start),
            LineFollowState::Run(start) if red => state = LineFollowState::Finish(start.elapsed()),
            _ => (),
        }
        buttons.process();
        // 500Hz update rate
        next += Duration::from_millis(2);
        sleep_until(next);
    }

    left.stop()?;
    right.stop()?;
    if let LineFollowState::Finish(time) = state {
        Ok(time)
    } else {
        Err(ProgramError::UserAbort)
    }
}

fn sleep_until(time: Instant) {
    sleep(time - Instant::now());
}

fn gray((r, g, b): (i32, i32, i32)) -> f32 {
    #[allow(clippy::cast_precision_loss)]
    let sum = (r + g + b) as f32;
    sum / 3.0
}

const fn is_red((r, g, b): (i32, i32, i32)) -> bool {
    let threshold = r / 2;
    // check for not black and green and blue significantly lower than red
    r > 32 && g < threshold && b < threshold
}
